# Copyright 2016 Kevin Murray <spam@kdmurray.id.au>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import bcolz
from docopt import docopt
from mpi4py import MPI
import numpy as np
import numexpr as ne
from pymer import CountMinKmerCounter

import itertools as itl
from glob import glob
import os
from os import path
import re
from sys import stderr, stdout

from .kernelmath import *
from .logging import *
from .utils import (
    calc_weights,
    count_reads,
    print_lsmat,
)


def count_main():
    cli = '''
    USAGE:
        kwipy-count [options] OUTFILE READFILES ...

    OPTIONS:
        -k KSIZE    Kmer length [default: 20]
        -N NTAB     Number of tables [default: 4]
        -x TSIZE    Table size [default: 1e9]

    Counts k-mers in READFILES to a count-min sketch which is saved to OUTFILE.
    '''

    opts = docopt(cli)
    k = int(opts['-k'])
    nt = int(opts['-N'])
    ts = int(float(opts['-x']))
    outfile = opts['OUTFILE']
    readfiles = opts['READFILES']

    counts = count_reads(*readfiles, k=k, sketchshape=(nt, ts))

    info("Writing counts to", outfile)
    counts.write(outfile)


def weight_main():
    cli = '''
    USAGE:
        kwipy-weight [options] WEIGHTFILE COUNTFILES ...

    OPTIONS:
        -k KSIZE    Kmer length [default: 20]
        -N NTAB     Number of tables [default: 4]
        -x TSIZE    Table size [default: 1e9]
    '''

    opts = docopt(cli)
    k = int(opts['-k'])
    nt = int(opts['-N'])
    ts = int(float(opts['-x']))
    outfile = opts['WEIGHTFILE']
    countfiles = opts['COUNTFILES']

    weights = calc_weights(*countfiles, k=k, sketchshape=(nt, ts))

    info("Writing weights to", outfile, end='... ')
    weights = bcolz.carray(popfreq, rootdir=outfile, mode='w')
    weights.flush()
    info("Done!")


def kernel_mpi_main():
    cli = '''
    USAGE:
        kwipy-kernel-mpi [options] OUTDIR WEIGHTFILE COUNTFILES ...

    OPTIONS:
        -c      Resume previous calculation to OUTDIR

    '''
    opts = docopt(cli)
    outdir = opts['OUTDIR']
    weightfile = opts['WEIGHTFILE']
    countfiles = opts['COUNTFILES']
    resume = opts['-c']

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    pairs = list(itl.combinations_with_replacement(countfiles, 2))

    if rank == 0:
        if resume:
            existing_kernlogs = glob(path.join(outdir, 'kernellog_*'))
            pairs_done = set()
            for kl in existing_kernlogs:
                with open(kl) as fh:
                    for line in fh:
                        a, b, kern = line.strip().split('\t')
                        pairs_done.add((a, b))
            pairs = pairs.filter(lambda x: x not in pairs_done)
        size = comm.Get_size()
        if size > len(pairs):
            warn('Number of MPI ranks is greater than number of comparisons')
            warn('This is harmless but silly')
        pieces = [list() for x in range(size)]
        for i, pair in enumerate(pairs):
            pieces[i%size].append(pair)
    else:
        pieces = None

    pairs = comm.scatter(pieces, root=0)

    weights = bcolz.open(weightfile, mode='r')

    outfile = path.join(outdir, 'kernellog_{}'.format(rank))
    if not resume:
        with open(outfile, 'w') as fh:
            pass

    for afile, bfile in pairs:
        a = bcolz.open(afile, mode='r')
        b = bcolz.open(bfile, mode='r')

        print(rank, afile, bfile)
        kernel = bcolz.eval('sum(a * b * weights, axis=0)').min()

        with open(outfile, 'a') as kfh:
            print(afile, bfile, kernel, sep='\t', file=kfh)


def distmat_main():
    cli = '''
    USAGE:
        kwipy-distmat [options] KERNELDIR

    OPTIONS:
        -k KERNEL   Output unnormalised kernel to KERNEL
        -d DIST     Output normalised distances to DIST
        -n NAME_RE  Regex to exract name from basename of count files. The
                    first match group is used. The regex must apply to all
                    files.  [default: (.*)\.(blz|kpy)]
        --keep      Keep kernel logs in KERNELDIR (which are normally removed)
    '''

    opts = docopt(cli)
    outdir = opts['KERNELDIR']
    kernfile = opts['-k']
    distfile = opts['-d']

    def _file_to_name(filename):
        filename = path.basename(filename)
        try:
            return re.search(opts['-n'], filename).groups()[0]
        except (TypeError, AttributeError):
            warn("Name regex did not match '{}', using basename"
                 .format(filename))
            return path.splitext(filename)[0]

    kernlogs = glob(path.join(outdir, 'kernellog_*'))
    kernels = {}
    samples = set()
    for klfile in kernlogs:
        with open(klfile) as fh:
            for line in fh:
                a, b, kern = line.strip().split('\t')
                a, b = map(_file_to_name, (a, b))
                kern = float(kern)
                samples.update((a, b))  # add a & b
                try:
                    kernels[a][b] = kern
                except KeyError:
                    kernels[a] = {b: kern}

    # Make an ordered array
    num_samples = len(samples)
    kernmat = np.zeros((num_samples, num_samples), dtype=float)
    samples = list(sorted(samples))
    sample_idx = {s:i for i, s in enumerate(samples)}
    # square up the dictionary
    for a, b in itl.combinations_with_replacement(sorted(samples), 2):
        ai = sample_idx[a]
        bi = sample_idx[b]
        kernmat[ai, bi] = kernels[a][b]
        kernmat[bi, ai] = kernels[a][b]

    if is_psd(kernmat):
        info("Kernel matrix is positive semi-definite")
    else:
        warn("Kernel matrix is NOT positive semi-definite")

    if is_psd(distmat):
        info("Distance matrix is positive semi-definite")
    else:
        warn("Distance matrix is NOT positive semi-definite")

    distmat = kernel_to_distance(kernmat)
    if distfile:
        with open(distfile, 'w') as dfh:
            print_lsmat(distmat, samples, file=dfh)

    if kernfile:
        with open(kernfile, 'w') as dfh:
            print_lsmat(kernmat, samples, file=dfh)
    else:
        print_lsmat(kernmat, samples)



import numpy as np
from numpy import linalg


def is_psd(matrix):
    '''Test if ``matrix`` is posititve semi-definite.'''
    v, w = linalg.eig(matrix)
    return (v > 0.0).all()


def normalise_kernel(K):
    '''Normalise a kernel matrix such that the diagonal is equal to 1.'''
    N = np.zeros_like(K)
    for i, j in itl.product(*map(range, K.shape)):
        N[i, j] = K[i, j] / np.sqrt(K[i,i] * K[j, j])
    return N


def kernel_to_distance(K):
    '''Convert a kernel matrix to a distance matrix.

    The kernel matrix need not be normalised, this is done internally.'''
    K = normalise_kern(kernel)
    D = np.zeros_like(kernel)
    for i, j in itl.product(*map(range, K.shape)):
        D[i,j] = (K[i,i] + K[j,j]) - 2*K[i,j]
    return D



# Copyright 2016 Kevin Murray <spam@kdmurray.id.au>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import bcolz
import numpy as np
from pymer import CountMinKmerCounter
import screed

from .logging import *


def print_lsmat(matrix, ids, file=None):
    print('', *ids, sep='\t', file=file)
    for rowidx, id in enumerate(ids):
        print(id, *list(matrix[rowidx,:]), sep='\t', file=file)


def count_reads(*readfiles, k=20, sketchshape=(4, 1e9)):
    counter = CountMinKmerCounter(k, sketchshape=sketchshape)

    for readfile in readfiles:
        info("Consuming:",  readfile)
        with screed.open(readfile) as reads:
            for i, read in enumerate(reads):
                counter.consume(str(read.sequence))
                if i % 10000 == 0:
                    progress(i/1000, 'K reads')
        progress(i/1000, 'K reads', end='\n')
    return counter


def calc_weights(*countfiles, k=20, sketchshape=(4, 1e9)):
    nt, ts = sketchshape
    popfreq = np.zeros((ts, nt), dtype=float)
    nsamples = len(countfiles)

    for countfile in countfiles:
        info("Loading",  countfile, end='... ')
        counts = bcolz.open(countfile, mode='r')
        # FIXME: the line below is due to a bug in bcolz, I think
        counts = np.array(counts).reshape(counts.shape)
        ne.evaluate('popfreq + where(counts > 0, 1, 0)', out=popfreq)
        info("Done!")

    info("Calculating entropy vector")

    # population counts to population freqs
    ne.evaluate('popfreq / nsamples', out=popfreq)
    # Population freqs to shannon entropy of pop freqs, i.e. weights
    ne.evaluate('-(popfreq * log(popfreq) + (1 - popfreq) * log((1-popfreq)))',
                out=popfreq)
    # workaround for numexpr's lack of isnan(): compare inequality, NaN != NaN
    ne.evaluate('where(popfreq != popfreq, 0, popfreq)', out=popfreq)



# Copyright 2016 Kevin Murray <spam@kdmurray.id.au>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

from blessings import Terminal

from sys import stdout, stderr


term = Terminal()


def progress(*args, file=stdout, end=''):
    file.write(term.move_x(0))
    file.write(term.clear_eol())
    file.write(term.cyan)
    print(*args, end=end, file=file)
    file.write(term.normal)
    file.flush()


def info(*args, file=stdout, end='\n'):
    file.write(term.blue)
    print(*args, file=file, end=end)
    file.write(term.normal)
    file.flush()


def warn(*args, file=stderr, end='\n'):
    file.write(term.bold_yellow)
    print(*args, file=file, end=end)
    file.write(term.normal)
    file.flush()



#!/usr/bin/env python

# Copyright  2016  Kevin Murray <spam@kdmurray.id.au>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from setuptools import setup
from setuptools.command.test import test as TestCommand
EXT='c'
try:
    from Cython.Build import cythonize
    EXT='pyx'
except ImportError:
    def cythonize(x): return x

import versioneer


class NoseCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly
        import nose
        nose.run_exit(argv=['nosetests'])


desc = """
kwip in python
"""

setup_requires = [
    'nose',
    'cython',
]

with open('requirements.txt') as fh:
    install_requires = [req.strip() for req in fh]

test_requires = [
    'coverage>=3.7',
]

command_classes=versioneer.get_cmdclass()
command_classes['test'] =  NoseCommand

setup(
    name="kwipy",
    packages=['kwipy', ],
    version=versioneer.get_version(),
    entry_points={
        'console_scripts': [
            'kwipy-count = kwipy.scripts:count_main',
            'kwipy-weight = kwipy.scripts:weight_main',
            'kwipy-kernel-mpi = kwipy.scripts:kernel_mpi_main',
            'kwipy-distmat = kwipy.scripts:distmat_main',
        ],
    },
    cmdclass=command_classes,
    install_requires=install_requires,
    tests_require=test_requires,
    setup_requires=setup_requires,
    description=desc,
    author="Kevin Murray",
    author_email="spam@kdmurray.id.au",
    url="https://github.com/kdmurray91/kwip",
    keywords=[
        "bioinformatics",
        "Next-gen Sequencing",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
    ],
)
